const BOOL = 1, INT = 2, CHAR = 3;

var FdExport = {
  DIRECTORY: "fastdial",
  PREFS_FILE: "fastdial.preferences",

  prefs: { scaleMethod: CHAR,
           width: INT,
           height: INT,
           thumbWidth: INT,
           fixedSize: BOOL,
           maxLoadTime: INT,
           shortcutKeys: CHAR,
           showAddMenu: BOOL,
           showInNewTabs: BOOL,
           showPopupButtons: BOOL,
           showIconAndTitle: BOOL,
           theme: CHAR },

  export: function() {
    var dir = FdExport.chooseDirectory();
    if (!dir) return;

    if (dir.leafName != FdExport.DIRECTORY) {
      dir.append(FdExport.DIRECTORY);
      if (!dir.exists()) {
        dir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, 0777);
      }
    }

    FdExport.exportPrefs(dir);

    // Copy image files

    var sourceDir = FdFile.getDataDirectory();
    var thumbsQty = FdPrefs.getIntPref("width") * FdPrefs.getIntPref("height");
    for(var i = 1; i <= thumbsQty; i++) {
      FdExport.copyImage(sourceDir, dir, FdFile.getImageName(i));
      FdExport.copyImage(sourceDir, dir, FdFile.getBigImageName(i));
    }
  },

  copyImage: function(sourceDir, targetDir, imageName) {
    var target = targetDir.clone();
    target.append(imageName);
    if (target.exists()) {
      target.remove(false);
    }
    var source = sourceDir.clone();
    source.append(imageName);
    if (source.exists()) {
      source.copyTo(targetDir, imageName);
    }
  },

  exportPrefs: function(dir) {
    var prefs = new Object();
    for(var name in FdExport.prefs) {
      var type = FdExport.prefs[name];
      var value = (type == CHAR) ? FdPrefs.getCharPref(name) :
                  (type == BOOL) ? FdPrefs.getBoolPref(name)
                                 : FdPrefs.getIntPref(name);
      prefs[name] = value;
    }
    var thumbsQty = prefs["width"] * prefs["height"];
    for(var i = 1; i <= thumbsQty; i++) {
      var thumb = FdPrefs.getCharPref(i);
      if (thumb) prefs[i] = thumb;
    }
    var prefsFile = FdExport.getPrefsFile(dir);
    FdFile.writeFile(prefsFile, FdUtils.toJSON(prefs, "\r\n"), true);
  },

  import: function() {
    var dir = FdExport.chooseDirectory();
    if (!dir) return;

    // Check if prefs file exists
    if (!FdExport.getPrefsFile(dir).exists()) {
      dir.append(FdExport.DIRECTORY);
      if (!dir.exists() || !FdExport.getPrefsFile(dir).exists()) {
        FdUtils.error("fileNotFound", [FdExport.PREFS_FILE]);
        return;
      }
    }

    FdExport.importPrefs(dir);

    //Copy image files
    var targetDir = FdFile.getDataDirectory();
    var thumbsQty = FdPrefs.getIntPref("width") * FdPrefs.getIntPref("height");
    for(var i = 1; i <= thumbsQty; i++) {
      FdExport.copyImage(dir, targetDir, FdFile.getImageName(i));
      FdExport.copyImage(dir, targetDir, FdFile.getBigImageName(i));
    }
    return true;
  },

  getPrefsFile: function(dir) {
    var file = dir.clone();
    file.append(FdExport.PREFS_FILE);
    return file;
  },

  importPrefs: function(dir) {
    var prefsFile = FdExport.getPrefsFile(dir);

    var prefs = eval(FdFile.read(prefsFile, true));
    for(var name in prefs) {
      var type = FdExport.prefs[name];
      var value = prefs[name];
      switch(type) {
        case CHAR: FdPrefs.setCharPref(name, value);
                   break;
        case BOOL: FdPrefs.setBoolPref(name, value);
                   break;
        case INT:  FdPrefs.setIntPref(name, value);
                   break;
      }
    }
    var thumbsQty = prefs["width"] * prefs["height"];
    for(var i = 1; i <= thumbsQty; i++) {
      if (prefs[i]) FdPrefs.setCharPref(i, prefs[i]);
      else FdPrefs.clearPref(i);
    }
  },

  chooseDirectory: function() {
    var fp = Components.classes["@mozilla.org/filepicker;1"]
             .createInstance(Components.interfaces.nsIFilePicker);
    fp.init(window, null, fp.modeGetFolder);
    if (fp.show() == fp.returnOK) return fp.file;
  }
}