var Fd = {
  IMAGE_BLANK: FdFile.getAsDataURL("chrome://fastdial/skin/blank.gif"),
  WIDTH: FdPrefs.getIntPref("width"),
  HEIGHT: FdPrefs.getIntPref("height"),
  currentId: 0,
  currentTarget: null,
  pageInfos: new Array(), // cached pageInfos
  prefsWindow: null,
  forceDrawPage: false,
  pageHtml: null,

  initialize: function() {
    var chromehidden = $("main-window").getAttribute("chromehidden");
    if (chromehidden.match(/location/)) return; // this is a popup window

    Fd.initShortcutKeys();

    window.addEventListener("resize", Fd.updatePage, false);
    gBrowser.addEventListener("load", Fd.onPageLoad, true);
    gBrowser.tabContainer.addEventListener("TabSelect", Fd.enableReloadButton, false);
    $("reload-button").addEventListener("click", Fd.onReload, false);
    $("contentAreaContextMenu").addEventListener("popupshowing",
                                                 Fd.onMainMenuShowing, false);

    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      if (Fd.pageInfos[i].refresh > 0) Fd.refreshThumbnail(i);
    }

    Fd.drawPagesOnStartup();
    Fd.showFF2notice();
  },

  enableReloadButton: function() {
    $("reload-button").disabled = false;
  },

  drawPagesOnStartup: function() {
    FdUtils.forEachBlankPage(function(doc) {
      var tab = FdUtils.getDocumentTab(doc);
      if (!tab.hasAttribute("busy") && FdUtils.isBlank(doc)) {
        Fd.drawPage(doc);
      }
    });
  },

  showFF2notice: function() {
    if (FdUtils.isFirefox2() &&
        FdPrefs.getCharPref("scaleMethod") == "default" &&
        FdPrefs.getBoolPref("showFF2notice"))
    {
      FdPrefs.setBoolPref("showFF2notice", false);
      openDialog("chrome://fastdial/content/ff2notice.xul",
               "fastdial-ff2notice-window",
               "chrome,centerscreen=yes");
    }
  },

  initShortcutKeys: function() {
    var modifiers = FdPrefs.getCharPref("shortcutKeys");
    for(var i = 1; i <= 9; i++) {
      var key = $("fastdial-key-" + i);
      key.setAttribute("modifiers", modifiers);
    }
  },

  open: function(event) {
    if (event.button == 0) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank");
    }
    else if (event.button == 1) {
      Fd.forceDrawPage = true;
      FdUtils.openUILink("about:blank", "tab");
    }
  },

  onPageLoad: function(event) {
    var doc = event.originalTarget;
    if (!FdUtils.getDocumentTab(doc) || // this is an iframe
        !FdUtils.isBlank(doc)) return;

    if (Fd.forceDrawPage ||
        gBrowser.browsers.length == 1 ||
        FdPrefs.getBoolPref("showInNewTabs"))
    {
      Fd.forceDrawPage = false;
      Fd.drawPage(doc);
    }
  },

  drawPage: function(doc) {
    if (!Fd.pageHtml) {
      Fd.pageHtml = Fd.createPageHtml();
    }
    doc.body.innerHTML = Fd.pageHtml;
    Fd.updateDoc(doc);

    doc.addEventListener("click", FdPage.onClick, false);
    doc.addEventListener("mousemove", FdPage.onMouseMove, false);
    doc.addEventListener("mousedown", FdPage.onMouseDown, false);
    doc.addEventListener("mouseup", FdPage.onMouseUp, false);
    doc.addEventListener("contextmenu", Fd.showMenu, false);

    var title = FdBundle.getString("fastDial");
    FdURL.addToHistory("about:blank", title);

    var showIconAndTitle = FdPrefs.getBoolPref("showIconAndTitle");
    doc.title = showIconAndTitle ? title : "";
    setTimeout(function() {
      var tab = FdUtils.getDocumentTab(doc);
      tab.setAttribute("image", showIconAndTitle
                                  ? "chrome://fastdial/skin/fastdial_icon.png" : "");
    }, 0);

    Fd.enableReloadButton();

    if (Fd.refreshAll) {
      for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
        Fd.refreshThumbnail(i);
      }
      Fd.refreshAll = false;
    }
  },

  invalidatePageHtml: function() {
    Fd.pageHtml = null;
  },

  createPageHtml: function() {
    Fd.WIDTH = FdPrefs.getIntPref("width");
    Fd.HEIGHT = FdPrefs.getIntPref("height");

    var hide = " style='display: none;'";

    var html = "";
    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      if (!Fd.pageInfos[i]) Fd.pageInfos[i] = FdPrefs.getPageInfo(i);
      var pageInfo = Fd.pageInfos[i];
      var title = pageInfo.title || "&nbsp;";

      if (i % Fd.WIDTH == 1) html += "<tr>";
      html += "<td><div class='cell' id='cell" + i + "'>" +
              "<div class='buttons' id='buttons" + i + "'>" +
              "<div class='button clear' title='" +
              FdBundle.getString("clear") + "'></div>" +
              "<div class='button refresh' title='" +
              FdBundle.getString("refresh") + "'></div>" +
              "<div class='button setURL' title='" +
              FdBundle.getString("setURL") + "'></div>" +
              "</div>" +
              "<div class='loading' id='loading" + i + "'" +
              (pageInfo._loading ? "" : hide) + "></div>" +
              "<a href='" + pageInfo.url + "' id='link" + i + "'" +
              (pageInfo._loading ? hide : "") + ">" +
              "<img class='image' id='" + i + "' src='" + Fd.getImageURL(i) + "'></a>" + 
              "<div class='title' id='title" + i + "'>" + title + "</div>" +
              "</div></td>";
      if (i % Fd.WIDTH == 0) html += "</tr>";
    }
    return FdTheme.current.template.replace(/%content%/, html);
  },

  openLink: function(event) {
    var id = FdPanel.getId(event.target);
    var pageInfo = Fd.pageInfos[id];
    FdUtils.openUILink(pageInfo.url, pageInfo.openIn || event);
  },

  exchangePanels: function(sourceId, targetId) {
    // Exchange pageInfos
    var targetPageInfo = Fd.pageInfos[sourceId];
    var sourcePageInfo = Fd.pageInfos[targetId];
    targetPageInfo.id = targetId;
    sourcePageInfo.id = sourceId;
    Fd.pageInfos[sourceId] = sourcePageInfo;
    Fd.pageInfos[targetId] = targetPageInfo;
    FdPrefs.savePageInfo(sourcePageInfo);
    FdPrefs.savePageInfo(targetPageInfo)

    // Exchange files
    FdFile.exchangeImages(sourceId, targetId);
    FdFile.exchangeBigImages(sourceId, targetId);

    // Update images
    Fd.updateImage(sourceId);
    Fd.updateImage(targetId);
  },

  showMenu: function(event) {
    Fd.currentId = FdPanel.getId(event.target);
    var pageInfo = Fd.currentId && Fd.pageInfos[Fd.currentId];

    $("fastdial-menu-setURL").hidden = !Fd.currentId;
    $("fastdial-menu-open-tab").hidden = !pageInfo || !pageInfo.url;
    $("fastdial-menu-open-tabshifted").hidden = !pageInfo || !pageInfo.url;
    $("fastdial-menu-open-separator").hidden = !pageInfo || !pageInfo.url;
    $("fastdial-menu-refresh").hidden = !pageInfo || !pageInfo.url;
    $("fastdial-menu-clear").hidden = !pageInfo || !pageInfo.url;
    $("fastdial-menu-share-logo").hidden = !pageInfo ||
                                           !pageInfo.customImage ||
                                           pageInfo.customImage.match(/^http:\/\/userlogos.org/);

    // Workaround for FF 2.x bug: otherwise popup is displayed in wrong place
    document.popupNode = null;

    var menu = $("fastdial-menu");
    menu.showPopup(menu, event.screenX, event.screenY, "context", null, null);

    event.stopPropagation();
    event.preventDefault();
  },

  // Gets id of the first empty cell

  getEmptyId: function() {
    for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
      if (!Fd.pageInfos[i].url) return i;
    }
  },

  onMainMenuShowing: function(event) {
    if (event.target.id == "contentAreaContextMenu") {
      var showMenu = FdPrefs.getBoolPref("showAddMenu");
      var showFullMenu = gContextMenu.target.src || !Fd.getEmptyId();
      $("fastdial-add-menu").hidden = !showMenu || showFullMenu;
      $("fastdial-add-full-menu").hidden = !showMenu || !showFullMenu;
      Fd.currentTarget = gContextMenu.target.href ||
                         gContextMenu.target.src;
    }
    else if (event.target.id == "fastdial-add-full-menu-popup") {
      FdUtils.clearElement(event.target);
      for(var i = 1; i <= Fd.WIDTH * Fd.HEIGHT; i++) {
        var menuitem = document.createElement("menuitem");
        menuitem.setAttribute("label", i + "." + Fd.pageInfos[i].title);
        menuitem.setAttribute("class", "menuitem-iconic");
        menuitem.setAttribute("image", FdURL.getFaviconURL(Fd.pageInfos[i].url));
        menuitem.setAttribute("oncommand", "Fd.addTarget(" + i + ");");
        event.target.appendChild(menuitem);
      }
    }
  },

  pageInfo: null,

  showSetURL: function(id) {
    var oldPageInfo = Fd.pageInfos[id];
    if (!Fd.pageInfo) Fd.pageInfo = FdUtils.clone(Fd.pageInfos[id]);

    openDialog("chrome://fastdial/content/seturl.xul",
               "fastdial-seturl-window",
               "chrome,centerscreen=yes,modal=yes,resizable=yes");

    if (Fd.pageInfo) {
      if (Fd.pageInfo.url) {

        Fd.pageInfos[id] = Fd.pageInfo;
        FdPrefs.savePageInfo(Fd.pageInfo);

        var doc = content.document;
        if (Fd.pageInfo.url != oldPageInfo.url || !Fd.pageInfo.title ||
            Fd.pageInfo.customImage != oldPageInfo.customImage ||
            Fd.pageInfo.customBackground != oldPageInfo.customBackground)
        {
          Fd.refreshThumbnail(id);
        }
        else {
          if (Fd.pageInfo.title != oldPageInfo.title) {
            Fd.updateImage(id);
          }
          if (Fd.pageInfo.refresh != oldPageInfo.refresh) {
            Fd.updateRefreshTimeout(id);
          }
        }
      } else Fd.clearURL(id, true);
    }
    Fd.pageInfo = null;
  },

  refreshThumbnail: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    pageInfo._loading = true;
    Fd.updateImage(id);

    if (pageInfo.customImage && pageInfo.title) {
      Fd.saveCustomImage(pageInfo);
    }
    else Fd.saveImage(pageInfo);
  },

  createFrame: function() {
    var browser = document.createElement("browser");
    browser.width = 1024;
    browser.height = 768;
    browser.setAttribute("type", "content");
    $("hidden-box").appendChild(browser);
    return browser;
  },

  deleteFrame: function(iframe) {
    iframe.parentNode.removeChild(iframe);
  },

  loadInFrame: function(url, onLoad) {
    function onFrameLoad() {
      iframe.removeEventListener("load", onFrameLoad, true);
      clearTimeout(loadTimeout);
      onLoad(iframe);
    }
    var iframe = Fd.createFrame();
    iframe.addEventListener("load", onFrameLoad, true);

    var loadTimeout = setTimeout(onFrameLoad,
                                 FdPrefs.getIntPref("maxLoadTime") * 1000);
    try {
      iframe.setAttribute("src", url);
    }
    catch(e) {
      FdUtils.error("incorrectURL", [url]);
      onFrameLoad();
    }
  },

  saveImage: function(pageInfo) {
    Fd.loadInFrame(pageInfo.url, function(iframe) {
        if (!pageInfo.title) {
          var doc = iframe.contentDocument;
          pageInfo.title = doc.title || doc.location.href;
          FdPrefs.savePageInfo(pageInfo);
        }
        if (pageInfo.customImage) {
          Fd.deleteFrame(iframe);
          Fd.saveCustomImage(pageInfo);
        }
        else Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveCustomImage: function(pageInfo) {
    var url = pageInfo.customImage + "#" + new Date().getTime(); // Workaround cache

    Fd.loadInFrame(url, function(iframe) {
        var doc = iframe.contentDocument;
        var img = doc.body.firstChild;
        if (FdThumbnail.getHeight(img.width) / img.height > 1) {
          doc.body.style.width = img.width;
          doc.body.style.height = FdThumbnail.getHeight(img.width);
        }
        else {
          doc.body.style.width = FdThumbnail.getWidth(img.height);
          doc.body.style.height = img.height;
        }
        doc.body.style.background = pageInfo.customBackground || "white";
        doc.body.style.margin = 0;
        doc.body.style.display = "table-cell";
        doc.body.style.textAlign = "center";
        doc.body.style.verticalAlign = "middle";

        Fd.saveThumbnail(pageInfo.id, iframe);
      });
  },

  saveThumbnail: function(id, iframe) {
    Fd.pageInfos[id]._loading = false;
    Fd.pageInfos[id]._imageURL = null;

    var thumbWidth =  FdPrefs.getIntPref("thumbWidth");
    var renderTimeout = FdPrefs.getIntPref("maxLoadTime") * 1000 / 5;

    setTimeout(function() {
        FdThumbnail.createThumbnail(iframe.contentWindow, thumbWidth,
          function(result) {
            Fd.deleteFrame(iframe);
            FdFile.saveBigImage(id, result && result.big);
            FdFile.saveImage(id, result && result.image);
            Fd.updateImage(id);
            Fd.updateRefreshTimeout(id);
          });
      },
      renderTimeout);
  },

  saveDoc: function(id, doc) {
    var iframe = Fd.createFrame();
    iframe.addEventListener("load", function() {
        Fd.saveThumbnail(id, iframe);
      },
      true);
    iframe.src = doc.location.href;
  },

  updateImage: function(id) {
    FdUtils.forEachBlankPage(function(doc) {
      var image = doc.getElementById(id);
      if (image) {
        var pageInfo = Fd.pageInfos[id];

        var url = Fd.getImageURL(id)
        FdURL.removeFromCache(url);
        image.src = url;

        var panel = new FdPanel(image);
        var link = panel.get("link");
        link.href = pageInfo.url;
        link.style.display = pageInfo._loading ? "none" : "block";

        panel.get("loading").style.display = pageInfo._loading ? "block" : "none";
        panel.get("title").innerHTML = pageInfo.title || "&nbsp;";
      }
    });
    Fd.invalidatePageHtml();
  },

  clearURL: function(id, dontAskConfirmation) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url) return;

    var title = pageInfo.title || pageInfo.url;
    var message = FdBundle.getString("clearConfirmation", [title]);
    if (dontAskConfirmation || FdUtils.confirm(message)) {
      clearTimeout(pageInfo._timeout);
      FdPrefs.clearPageInfo(id);
      Fd.pageInfos[id] = FdPrefs.getPageInfo(id); // update pageInfo cache
      FdFile.deleteImage(id);
      FdFile.deleteBigImage(id);
      Fd.updateImage(id);
    }
  },

  getImageSize: function() {
    var width, height;
    var style = FdTheme.current.style;

    if (FdPrefs.getBoolPref("fixedSize")) {
      width = FdPrefs.getIntPref("thumbWidth");
      height = FdThumbnail.getHeight(width);
    } else {
      width = (content.innerWidth - (style.body.margin*2 + style.content.borderSpacing)) /
              Fd.WIDTH - (parseInt(style.cell.border)*2 + style.cell.padding*2 +
              parseInt(style.image.border)*2 + style.content.borderSpacing);

      height = (content.innerHeight - (style.body.margin*2 + style.content.borderSpacing)) /
               Fd.HEIGHT - (parseInt(style.cell.border)*2 + style.cell.padding +
               parseInt(style.image.border)*2 + style.content.borderSpacing +
               style.title.margin*2 + style.body.fontSize + 6);
    }
    width = Math.max(Math.floor(width), style.image.minWidth);
    height = Math.max(Math.floor(height), style.image.minHeight);

    if (FdThumbnail.getHeight(width) / height > 1) {
      width = FdThumbnail.getWidth(height);
    }
    else height = FdThumbnail.getHeight(width);

    return { width: width, height: height };
  },

  updatePage: function() {
    FdUtils.forEachBlankPage(Fd.updateDoc);
  },

  redrawPage: function() {
    FdUtils.forEachBlankPage(Fd.drawPage);
  },

  updateDoc: function(doc) {
    var imageSize = Fd.getImageSize();
    var style = FdTheme.current.style;

    for(var i = 0; i < doc.styleSheets.length; i++) {
      var styleSheet = doc.styleSheets[i];
      if (styleSheet.href &&
          styleSheet.href != "about:blank") continue;

      var rules = styleSheet.cssRules;
      for(var j = 0; j < rules.length; j++) {
        var rule = rules[j];
        switch(rule.selectorText) {
          case ".loading":
          case ".image":
              rule.style.height = imageSize.height;
              rule.style.width = imageSize.width;
              break;
          case ".title":
              rule.style.width = imageSize.width +
                                 parseInt(style.image.border)*2 - style.title.margin*2;
              break;
        }
        Fd.applyStyle(rule, style);
      }
    }
  },

  applyStyle: function(rule, style) {
    if (!rule.selectorText) return;

    var regexp = /\s*([^#.]+)\s*(,|$)/g;
    while(match = regexp.exec(rule.selectorText)) {
      var entry = style[match[1]];
      if (!entry) continue;

      if (rule.selectorText.match(/hover/)) {
        entry = entry.hover;
      }
      for(var i in entry) {
        rule.style[i] = entry[i];
      }
    }
  },

  openPreferences: function() {
    if (!Fd.prefsWindow || Fd.prefsWindow.closed) {
      Fd.prefsWindow = window.openDialog("chrome://fastdial/content/preferences.xul",
                                         "fastdial-preferences-window",
                                         "chrome,toolbar,centerscreen,resizable=yes");
    } else Fd.prefsWindow.focus();
  },

  getImageURL: function(id) {
    var pageInfo = Fd.pageInfos[id];

    if (pageInfo.url && pageInfo.imageURL == null) {
      pageInfo._imageURL = FdFile.getImageURL(id);
    }
    var url = pageInfo._imageURL || Fd.IMAGE_BLANK;
    return url;
  },

  openPage: function(id, where) {
    var pageInfo = FdPrefs.getPageInfo(id);
    FdUtils.openUILink(pageInfo.url, where);
  },

  updateRefreshTimeout: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (pageInfo._timeout) clearTimeout(pageInfo._timeout);
    if (pageInfo.refresh > 0) {
      pageInfo._timeout = setTimeout(function() { Fd.refreshThumbnail(pageInfo.id); },
                                     pageInfo.refresh * 60 * 1000);
    }
  },

  // For custom images show "Set URL" dialog, for site URLs - silently add

  addTarget: function(id) {
    if (!id) id = Fd.getEmptyId();
    var customImage = FdURL.isImage(Fd.currentTarget) && Fd.currentTarget;
    if (customImage) {
      Fd.pageInfo = FdUtils.clone(Fd.pageInfos[id]);
      Fd.pageInfo.customImage = customImage;
    } else {
      Fd.pageInfo = new FdPageInfo(id);
      Fd.pageInfo.url = Fd.currentTarget || content.document.location.href;
      if (!Fd.currentTarget) Fd.pageInfo.title = content.document.title;
    }
    Fd.showSetURL(id);
  },

  // We don't want to reload the page itself, just want to reload all thumbnails. But
  // event.stopPropagation() & event.preventDefault() methods don't stop reload event,
  // so just defer reloading of thumbnails till the page itself is reloaded

  onReload: function(event) {
    if (FdUtils.isBlank(content.document)) {
      if (!FdPrefs.getBoolPref("confirmRefreshAll")) Fd.refreshAll = true;
      else {
        var dontAsk = { value: false };
        Fd.refreshAll = FdUtils.confirmCheck(FdBundle.getString("refreshAll"),
                                             FdBundle.getString("dontAskNextTime"),
                                             dontAsk);
        FdPrefs.setBoolPref("confirmRefreshAll", !dontAsk.value);
      }
    }
  },

  findLogo: function() {
    var url = Fd.currentId && Fd.pageInfos[Fd.currentId].url;
    if (url) {
      var host = FdURL.getNsiURL(url).host;
      FdUtils.openUILink("http://userlogos.org/search/node/" +
                         encodeURIComponent(host + " type:logo"));
    }
    else FdUtils.openUILink("http://userlogos.org/");
  },

  shareLogo: function() {
    FdUtils.openUILink("http://userlogos.org/node/add/logo");
  },

  showZoom: function(id) {
    var pageInfo = Fd.pageInfos[id];
    if (!pageInfo.url || !FdFile.getBigImageFile(id).exists()) return;

    var doc = content.document;
    var zoom = doc.createElement("div");
    zoom.id = "zoom";
    zoom.innerHTML = "<div id='zoom-background'></div>" + 
                     "<div id='zoom-image'>" +
                     "<img src='" + FdFile.getBigImageURL(id) + "'></div>";
    doc.body.appendChild(zoom);

    function onMouseUp(event) {
      var zoom = doc.getElementById("zoom");
      zoom.parentNode.removeChild(zoom);
      doc.removeEventListener("mouseup", onMouseUp, true);
      event.preventDefault();
    }
    doc.addEventListener("mouseup", onMouseUp, true);
  }
}
