function FdPanel(element) {
  this.doc = element.ownerDocument;
  this.id = FdPanel.getId(element);

  this.get = function(type) {
    return this.id ? this.doc.getElementById(type + this.id) : null;
  }
}
FdPanel.getId = function(element) {
  if (element.className.match(/button\b/)) {
    element = element.parentNode;
  }
  var match = element.id.match(/\d+$/);
  return match && match[0];
}

function FdPageImpl() {
  const TIMEOUT_SHOW = 300;
  const TIMEOUT_HIDE = 300;

  var x, y;
  var clickX, clickY;
  var hoverButtons, visibleButtons = new Array();
  var drag = { source: null, x: 0, y: 0, object: null };
  var zoomTimeout;

  function getAbsolutePosition(element) {
    var panel = new FdPanel(element);
    var cell = panel.get("cell") || {};
    return { x: cell.offsetLeft + element.offsetLeft,
             y: cell.offsetTop + element.offsetTop,
             width: element.offsetWidth,
             height: element.offsetHeight };
  }

  function isMouseOver(element) {
    var position = getAbsolutePosition(element);
    return (x >= position.x && x <= position.x + position.width &&
            y >= position.y && y <= position.y + position.height);
  }

  function showButtons(buttons) {
    var vb = visibleButtons[buttons.id];
    if (vb) {
      if (vb.hideTimeout) {
        clearTimeout(vb.hideTimeout);
        vb.hideTimeout = null;
      }
    }
    else if (buttons != hoverButtons) {
      if (hoverButtons) {
        clearTimeout(hoverButtons.showTimeout);
      }
      hoverButtons = buttons;
      hoverButtons.showTimeout = setTimeout(doShowButtons, TIMEOUT_SHOW);
    }
  }

  function doShowButtons() {
    if (isMouseOver(hoverButtons)) {
      hoverButtons.style.zIndex = 0;
      visibleButtons[hoverButtons.id] = hoverButtons;
    }
  }

  function hideButtons() {
    if (hoverButtons) {
      clearTimeout(hoverButtons.showTimeout);
      hoverButtons = null;
    }

    function createHideFunction(id) {
      return function() { doHideButtons(id); }
    }

    for(var id in visibleButtons) {
      if (!visibleButtons[id].hideTimeout) {
        visibleButtons[id].hideTimeout =
          setTimeout(createHideFunction(id), TIMEOUT_HIDE);
      }
    }
  }

  function doHideButtons(id) {
    var buttons = visibleButtons[id];
    buttons.style.zIndex = -1;
    buttons.hideTimeout = null;
    delete visibleButtons[id];
  }

  this.onMouseMove = function(event) {
    x = event.pageX;
    y = event.pageY;

    if (drag.source && !drag.object &&
        Math.abs(clickX - x) + Math.abs(clickY - y) > 10) // minumum drag interval
    {
      drag.object = drag.source.cloneNode(true);
      drag.source.src = Fd.IMAGE_BLANK;
      drag.object.style.position = "absolute";
      var doc = event.target.ownerDocument;
      doc.body.appendChild(drag.object);
    }
    if (drag.object) {
      drag.object.style.left = event.pageX - drag.x;
      drag.object.style.top = event.pageY - drag.y;
      drag.object.style.cursor = getHoverImage(event) ? "move" : "not-allowed";
    }
    else if (FdPrefs.getBoolPref("showPopupButtons")) {
      var panel = new FdPanel(event.target);
      if (panel.id) {
        var buttons = panel.get("buttons");
        if (isMouseOver(buttons)) {
          return showButtons(buttons);
        }
      }
    }
    hideButtons();
  }

  this.onMouseDown = function(event) {
    clickX = event.pageX;
    clickY = event.pageY;

    if (drag.object) {
      // If we got here, then user has released dragged object outside
      // browser window and we missed onMouseUp event. Perform it now.
      FdPage.onMouseUp(event);
    }
    else if (event.target.nodeName == "IMG") {
      if (event.button == 0) {
        drag.source = event.target;
        var position = getAbsolutePosition(event.target);
        drag.x = clickX - position.x;
        drag.y = clickY - position.y;
        event.preventDefault();
      }
      else if (event.button == 2) {
        zoomTimeout = setTimeout(function() {
          Fd.showZoom(event.target.id);
        }, 200); // 0.2 seconds
      }
    }
  }

  function getHoverImage(event) {
    var doc = event.target.ownerDocument;
    var image, i = 1;
    while(image = doc.getElementById(i++)) {
      if (isMouseOver(image)) return image;
    }
  }

  this.onMouseUp = function(event) {
    clearTimeout(zoomTimeout);

    if (drag.object) {
      var doc = event.target.ownerDocument;
      doc.body.removeChild(drag.object);

      var target = getHoverImage(event);
      if (target && target != drag.source) {
        Fd.exchangePanels(drag.source.id, target.id);
      } else {
        // Return image back to its place
        drag.source.src = drag.object.src;
      }
      drag.object = null;
      event.preventDefault();
    }
    drag.source = null;
  }

  this.onClick = function(event) {
    var panel = new FdPanel(event.target);
    if (!panel.id || event.button == 2) return;

    event.stopPropagation();
    event.preventDefault();

    if (event.button == 0) {
      switch(event.target.className) {
        case "button clear":   return Fd.clearURL(panel.id);
        case "button refresh": return Fd.refreshThumbnail(panel.id);
        case "button setURL":  return Fd.showSetURL(panel.id);
      }
      if (event.target.className != "loading" &&
          event.target.className != "buttons" &&
          (event.target.nodeName != "IMG" || !panel.get("link").href))
      {
        return Fd.showSetURL(panel.id);
      }
    }
    if (event.button == 1) {
      if (event.target.className == "title" ||
          event.target.className == "cell")
      {
        return Fd.refreshThumbnail(panel.id);
      }
    }
    if (!event.target.className.match(/button/)) {
      Fd.openLink(event);
    }
  }
}

var FdPage = new FdPageImpl();
